/* ================================================================================ */
/* |                                                                              | */
/* |         logisticCIC[@ŉvO                       | */
/* |                                                                              | */
/* |         x_{t}  =  x * ( 1-x )                                                | */
/* |                                                                              | */
/* |                                       This program is in the public domain.  | */
/* |                                Programmed by M.Yadome. 2008/12/18, ver. 1.0  | */
/* |                                                                              | */
/* ================================================================================ */


/* ================================================================================ */
/*           wb_t@C                                                         */
/* ================================================================================ */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>


/* ================================================================================ */
/*           }N`                                                             */
/* ================================================================================ */
#define  DATADIR   "data"         /*  f[^fBNg                          */
#define  DIRMODE   0744           /*  f[^fBNg̋[h            */
#define  FILENAME  "ode1em.dat"   /*  f[^t@C                              */
#define  MAXSTEP   10000          /*  ővZ                                  */
#define  SOUTPUT   100            /*  SOUTPUTƂɃf[^o                       */
#define  SLEEP     0.01           /*  SLEEPb~                         */
#define  dt        0.01           /*  ԍݕ                                    */


/* ================================================================================ */
/*           logistic equation                                                      */
/* ================================================================================ */
double F( double x ){   return x*(1-x);   }


/* ================================================================================ */
/*           f[^fBNg̍쐬                                               */
/* ================================================================================ */
void make_dir()
{
  char str[100];
  if(  mkdir(DATADIR,DIRMODE) != 0  ){
    sprintf( str, "rm -rf %s", DATADIR );
    system( str );
    mkdir( DATADIR, DIRMODE );
  }
}


/* ================================================================================ */
/*           t@C݃[hŊJ                                         */
/* ================================================================================ */
FILE *fopen_w( char filename[] )
{
  FILE *fp;
  if(  (fp=fopen(filename,"w")) == NULL  ){
    printf( "cannot open output file\n" );
    exit(1);
  }
  return fp;
}


/* ================================================================================ */
/*           f[^t@C̍쐬                                                 */
/* ================================================================================ */
FILE *output_filename()
{
  char filename[100];
  FILE *fp;
  sprintf( filename, "%s/%s", DATADIR, FILENAME );
  fp  =  fopen_w( filename );
  return fp;
}


/* ================================================================================ */
/*           gnuplot̏ݒ                                                      */
/* ================================================================================ */
FILE *gnuplot_ini()
{
  FILE *gp=NULL;
  gp  =  popen( "gnuplot -geometry 400x400+40+50" , "w" );
  fprintf( gp, "set xrange[0:%lf] \n", MAXSTEP*dt );  
  return gp;
}


/* ================================================================================ */
/*           l̐ݒ                                                           */
/* ================================================================================ */
void initial_condition( double *x )
{
  char c;
  printf( "input initial data [Y/N]:" );
  scanf( "%c%*c", &c );
  if(  c == 'y'  ){
    printf( "x  =  " );  scanf( "%lf%*c", x );
  }else{
    *x  =  0.1;
  }
}


/* ================================================================================ */
/*           f[^̏o                                                           */
/* ================================================================================ */
void output( FILE *fp, double t, double x )
{
  fprintf( fp, "%20.15lf %20.15lf\n", t, x );
  fflush( fp );
}


/* ================================================================================ */
/*           CIC[@                                                         */
/* ================================================================================ */
void modified_euler_method( double *xx )
{
  double x;
  double kx;
  x    =  *xx;
  kx   =  F( x );
  *xx  =  x+dt*F( x+dt*kx/2.0 );
}


/* ================================================================================ */
/*           gnuplot̏o                                                          */
/* ================================================================================ */
void gnuplot1d( FILE *fp )
{
  fprintf( fp, "plot \"%s/%s\" u 1:2 w l\n", DATADIR, FILENAME );
  fflush( fp );
}


/* ================================================================================ */
/*           main֐                                                               */
/* ================================================================================ */
int main()                                    /*                                    */
{                                             /*                                    */
  int    step;                                /*                                    */
  FILE   *fp, *gp;                            /*                                    */
  double x;                                   /*                                    */
  make_dir();                                 /*  f[^fBNg̍쐬          */
  step  =  0;                                 /*                                    */
  initial_condition( &x );                    /*  l̐ݒ                      */
  fp  =  output_filename();                   /*  f[^t@C̍쐬            */
  gp  =  gnuplot_ini();                       /*  gnuplot̏ݒ                 */
  output( fp, step*dt, x );                   /*  f[^̏o                      */
  gnuplot1d( gp );                            /*  gnuplot̏o                     */
  usleep( SLEEP*1.0E+06 );                    /*  SLEEPb~             */
  for( step = 1; step <= MAXSTEP; step++ ){   /*                                    */
    modified_euler_method( &x );              /*  CIC[@                    */
    if(  step%SOUTPUT == 0  ){                /*                                    */
      output( fp, step*dt, x );               /*  f[^̏o                      */
      gnuplot1d( gp );                        /*  gnuplot̏o                     */
      usleep( SLEEP*1.0E+06 );                /*  SLEEPb~             */
    }                                         /*                                    */
  }                                           /*                                    */
  printf( "please hit return key!!" );        /*                                    */
  getchar();                                  /*  W͂Pǂݍ      */
  fclose( fp );                               /*  t@C                  */
  pclose( gp );                               /*  gnuplot                   */
  return 0;                                   /*  I                          */
}                                             /*                                    */
