/* ================================================================================ */
/* |                                                                              | */
/* |         LorenzS̃QENb^@ŉvO                 | */
/* |                                                                              | */
/* |         x_{t}  =  rho * ( y - x )                                            | */
/* |         y_{t}  =  -y -x*z +r*x                                               | */
/* |         z_{t}  =  x*y - b*z                                                  | */
/* |                                                                              | */
/* |                                       This program is in the public domain.  | */
/* |                                Programmed by M.Yadome. 2008/12/18, ver. 1.0  | */
/* |                                                                              | */
/* ================================================================================ */


/* ================================================================================ */
/*           wb_t@C                                                         */
/* ================================================================================ */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>


/* ================================================================================ */
/*           }N`                                                             */
/* ================================================================================ */
#define  DATADIR   "data"         /*  f[^fBNg                          */
#define  DIRMODE   0744           /*  f[^fBNg̋[h            */
#define  FILENAME  "ode3rk.dat"   /*  f[^t@C                              */
#define  MAXSTEP   200000         /*  ővZ                                  */
#define  SOUTPUT   160            /*  SOUTPUTƂɃf[^o                       */
#define  SLEEP     0.02           /*  SLEEPb~                         */
#define  dt        0.0002         /*  ԍݕ                                    */
#define  rho       10.0           /*  ̃p[^ݒ                        */
#define  r         26.0           /*  ̃p[^ݒ                        */
#define  b         8.0/3.0        /*  ̃p[^ݒ                        */


/* ================================================================================ */
/*           Lorenz equation                                                        */
/* ================================================================================ */
double F( double x, double y, double z ){   return rho*(y-x);    }
double G( double x, double y, double z ){   return -y-x*z+r*x;   }
double H( double x, double y, double z ){   return x*y-b*z;      }


/* ================================================================================ */
/*           f[^fBNg̍쐬                                               */
/* ================================================================================ */
void make_dir()
{
  char str[100];
  if(  mkdir(DATADIR,DIRMODE) != 0  ){
    sprintf( str, "rm -rf %s", DATADIR );
    system( str );
    mkdir( DATADIR, DIRMODE );
  }
}


/* ================================================================================ */
/*           t@C݃[hŊJ                                         */
/* ================================================================================ */
FILE *fopen_w( char filename[] )
{
  FILE *fp;
  if(  (fp=fopen(filename,"w")) == NULL  ){
    printf( "cannot open output file\n" );
    exit(1);
  }
  return fp;
}


/* ================================================================================ */
/*           f[^t@C̍쐬                                                 */
/* ================================================================================ */
FILE *output_filename()
{
  char filename[100];
  FILE *fp;
  sprintf( filename, "%s/%s", DATADIR, FILENAME );
  fp  =  fopen_w( filename );
  return fp;
}


/* ================================================================================ */
/*           gnuplot̏ݒ                                                      */
/* ================================================================================ */
FILE *gnuplot_ini( int select )
{
  FILE *gp=NULL;
  if(  select == 1  ){
    gp = popen( "gnuplot -geometry 400x400+40+50" , "w" );
    fprintf( gp, "set xrange[0:%lf] \n", MAXSTEP*dt );  
  }else if(  select == 2  )
    gp = popen( "gnuplot -geometry 400x400+450+50", "w" );
  return gp;
}


/* ================================================================================ */
/*           l̐ݒ                                                           */
/* ================================================================================ */
void initial_condition( double *x, double *y, double *z )
{
  char c;
  printf( "input initial data [Y/N]:" );
  scanf( "%c%*c", &c );
  if(  c == 'y'  ){
    printf( "x  =  " );  scanf( "%lf%*c", x );
    printf( "y  =  " );  scanf( "%lf%*c", y );
    printf( "z  =  " );  scanf( "%lf%*c", z );
  }else{
    *x  =  0.1;
    *y  =  0.1;
    *z  =  0.1;
  }
}


/* ================================================================================ */
/*           f[^̏o                                                           */
/* ================================================================================ */
void output( FILE *fp, double t, double x, double y, double z )
{
  fprintf( fp, "%20.15lf %20.15lf %20.15lf %20.15lf\n", t, x, y, z );
  fflush( fp );
}


/* ================================================================================ */
/*           S̃QENb^@                                                 */
/* ================================================================================ */
void runge_kutta_4( double *xx, double *yy, double *zz )
{
  double x  , y  , z;
  double kx1, ky1, kz1;
  double kx2, ky2, kz2;
  double kx3, ky3, kz3;
  double kx4, ky4, kz4;
  x    =  *xx;
  y    =  *yy;
  z    =  *zz;
  kx1  =  F( x           , y           , z            );
  ky1  =  G( x           , y           , z            );
  kz1  =  H( x           , y           , z            );
  kx2  =  F( x+dt*kx1/2.0, y+dt*ky1/2.0, z+dt*kz1/2.0 );
  ky2  =  G( x+dt*kx1/2.0, y+dt*ky1/2.0, z+dt*kz1/2.0 );
  kz2  =  H( x+dt*kx1/2.0, y+dt*ky1/2.0, z+dt*kz1/2.0 );
  kx3  =  F( x+dt*kx2/2.0, y+dt*ky2/2.0, z+dt*kz2/2.0 );
  ky3  =  G( x+dt*kx2/2.0, y+dt*ky2/2.0, z+dt*kz2/2.0 );
  kz3  =  H( x+dt*kx2/2.0, y+dt*ky2/2.0, z+dt*kz2/2.0 );
  kx4  =  F( x+dt*kx3    , y+dt*ky3    , z+dt*kz3     );
  ky4  =  G( x+dt*kx3    , y+dt*ky3    , z+dt*kz3     );
  kz4  =  H( x+dt*kx3    , y+dt*ky3    , z+dt*kz3     );
  *xx  =  x + dt * ( kx1 + 2.0*kx2 + 2.0*kx3 + kx4 ) / 6.0;           
  *yy  =  y + dt * ( ky1 + 2.0*ky2 + 2.0*ky3 + ky4 ) / 6.0;
  *zz  =  z + dt * ( kz1 + 2.0*kz2 + 2.0*kz3 + kz4 ) / 6.0;
}


/* ================================================================================ */
/*           gnuplot̏o                                                          */
/* ================================================================================ */
void gnuplot3d( FILE *fp2, FILE *fp3 )
{
  fprintf( fp2, "plot  \"%s/%s\" u 1:2 w l, \"\" u 1:3 w l, \"\" u 1:4 w l\n", DATADIR, FILENAME );
  fprintf( fp3, "splot \"%s/%s\" u 2:3:4 w l\n"                              , DATADIR, FILENAME );
  fflush( fp2 );
  fflush( fp3 );
}


/* ================================================================================ */
/*           main֐                                                               */
/* ================================================================================ */
int main()                                    /*                                    */
{                                             /*                                    */
  int    step;                                /*                                    */
  FILE   *fp, *gp1, *gp2;                     /*                                    */
  double x, y, z;                             /*                                    */
  make_dir();                                 /*  f[^fBNg̍쐬          */
  step  =  0;                                 /*                                    */
  initial_condition( &x, &y, &z );            /*  l̐ݒ                      */
  fp   =  output_filename();                  /*  f[^t@C̍쐬            */
  gp1  =  gnuplot_ini( 1 );                   /*  gnuplot̏ݒ                 */
  gp2  =  gnuplot_ini( 2 );                   /*  gnuplot̏ݒ                 */
  output( fp, step*dt, x, y, z );             /*  f[^̏o                      */
  gnuplot3d( gp1, gp2 );                      /*  gnuplot̏o                     */
  usleep( SLEEP*1.0E+06 );                    /*  SLEEPb~             */
  for( step = 1; step <= MAXSTEP; step++ ){   /*                                    */
    runge_kutta_4( &x, &y, &z );             /*  S̃QENb^@            */
    if(  step%SOUTPUT == 0  ){                /*                                    */
      output( fp, step*dt, x, y, z );         /*  f[^̏o                      */
      gnuplot3d( gp1, gp2 );                  /*  gnuplot̏o                     */
      usleep( SLEEP*1.0E+06 );                /*  SLEEPb~             */
    }                                         /*                                    */
  }                                           /*                                    */
  printf( "please hit return key!!" );        /*                                    */
  getchar();                                  /*  W͂Pǂݍ      */
  fclose( fp  );                              /*  t@C                  */
  pclose( gp1 );                              /*  gnuplot                   */
  pclose( gp2 );                              /*  gnuplot                   */
  return 0;                                   /*  I                          */
}                                             /*                                    */
