/* ================================================================================ */
/* |                                                                              | */
/* |         gFitzHugh-Nagumo                                      | */
/* |         NNjR\@ŉvO - period -                      | */
/* |                                                                              | */
/* |         u_{t}  =  du * u_{xx} + (u*(1-u)*(u-a)-v) /eps                       | */
/* |         v_{t}  =  dv * v_{xx} +  x-gam*y                                     | */
/* |                                                                              | */
/* |                                       This program is in the public domain.  | */
/* |                                Programmed by M.Yadome. 2008/12/18, ver. 1.0  | */
/* |                                                                              | */
/* ================================================================================ */


/* ================================================================================ */
/*           wb_t@C                                                         */
/* ================================================================================ */
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>


/* ================================================================================ */
/*           }N`                                                             */
/* ================================================================================ */
#define  DATADIR  "data"                   /*  f[^fBNg                 */
#define  DIRMODE  0744                     /*  f[^fBNg̋[h   */
#define  MAXSTEP  15000                    /*  ővZ                         */
#define  SOUTPUT  100                      /*  SOUTPUTƂɃf[^o              */
#define  SLEEP    0.0                      /*  SLEEPb~                */
#define  dt       0.0005                   /*  ԍݕ                           */
#define  N        400                      /*  ԕ                           */
#define  Lx       80.0                     /*                                   */
#define  du       1.0                      /*  gUW                             */
#define  dv       0.1                      /*  gUW                             */
#define  eps      0.001                    /*  ̃p[^                   */
#define  aa       0.125                    /*  ̃p[^                   */
#define  gam      1.0                      /*  ̃p[^                   */
#define  BCL      0.0                      /*  E()                         */
#define  BCR      0.0                      /*  E(E)                         */
#define  theta    0.5                      /*  z@ƉA@̔䗦                 */
#define  dx       Lx/N
#define  re       (1.0-theta)*dt/(dx*dx)
#define  ri       (    theta)*dt/(dx*dx)
#define  M        N-1


/* ================================================================================ */
/*           f[^fBNg̍쐬                                               */
/* ================================================================================ */
void make_dir()
{
  char str[100];
  if(  mkdir(DATADIR,DIRMODE) != 0  ){
    sprintf( str, "rm -rf %s", DATADIR );
    system( str );
    mkdir( DATADIR, DIRMODE );
  }
}


/* ================================================================================ */
/*           l̐ݒ                                                           */
/* ================================================================================ */
void initial_condition( double u[], double v[] )
{
  int i;
  for( i = 1; i <= N; i++ ){
    u[i]  =  0.0;
    v[i]  =  0.0;
    if( N/5 <= i  &&  i <= (2*N)/5 )   u[i] = 1.0;
    if( 1   <= i  &&  i <=     N/5 )   v[i] = 0.2;
  }
}


/* ================================================================================ */
/*           FitzHugh-Nagumo equation                                               */
/* ================================================================================ */
void reaction_term( double u[], double v[], double f[], double g[] )
{
  int i;
  for( i = 0; i <= M; i++ ){
    f[i]  =  (u[i]*(1.0-u[i])*(u[i]-aa)-v[i])/eps;
    g[i]  =  u[i]-gam*v[i];
  }
}


/* ================================================================================ */
/*                                                                              */
/* ================================================================================ */
int digit_number( int num )
{
  int dn;
  dn  =  1;
  while( (num=num/10)>0 ){   dn++;   }
  return dn;
}


/* ================================================================================ */
/*           t@C݃[hŊJ                                         */
/* ================================================================================ */
FILE *fopen_w( char filename[] )
{
  FILE *fp;
  if(  (fp=fopen(filename,"w")) == NULL  ){
    printf( "cannot open output file\n" );
    exit(1);
  }
  return fp;
}


/* ================================================================================ */
/*           f[^̏o                                                           */
/* ================================================================================ */
FILE *output_gnuplot_ini()
{
  char filename[100];
  FILE *fp;
  sprintf( filename, "%s/plot", DATADIR );
  fp  =  fopen_w( filename );
  fprintf( fp, "set xrange[0:%20.15lf]\n", Lx );
  fprintf( fp, "set yrange[-0.4:1.2]\n"       );  
  return fp;
}
void output_gnuplot( int number, FILE *fp )
{
  fprintf( fp, "!echo \"%0*d\"\n"                         , digit_number(MAXSTEP/SOUTPUT), number );
  fprintf( fp, "plot \"%0*d\" u 2:3 w l, \"\" u 2:4 w l\n", digit_number(MAXSTEP/SOUTPUT), number );
  fprintf( fp, "!sleep %lf\n"                             , SLEEP                                 );
  fflush( fp );
}
void output_data( double time, double u[], double v[], FILE *fp )
{
  int i;
  for( i = 0; i <= N; i++ )
    fprintf( fp, "%20.15E %20.15E %20.15E %20.15E\n", time, i*dx, u[i], v[i] );
  fflush( fp );
}
void output( int number, double time, double u[], double v[], FILE *fp_gnuplot )
{
  char filename[100];
  FILE *fp_data;
  sprintf( filename, "%s/%0*d", DATADIR, digit_number(MAXSTEP/SOUTPUT), number );
  printf( "%s\n", filename );
  fp_data  =  fopen_w( filename );
  output_data( time, u, v, fp_data );
  fclose( fp_data );
  output_gnuplot( number, fp_gnuplot );
}


/* ================================================================================ */
/*           gnuplot̏ݒƏo                                                */
/* ================================================================================ */
FILE *gnuplot_ini()
{
  FILE *fp;
  fp  =  popen( "gnuplot -geometry 400x400+40+50", "w" );
  fprintf( fp, "set xrange[0:%20.15lf]\n", Lx );
  fprintf( fp, "set yrange[-0.4:1.2]\n"       );  
  return fp;
}
void gnuplot( int number, FILE *gp )
{
  fprintf( gp, "plot \"%s/%0*d\" u 2:3 w l, \"\" u 2:4 w l\n", DATADIR, digit_number(MAXSTEP/SOUTPUT), number );
  fflush( gp );
}


/* ================================================================================ */
/*           LUpĘAP莮                                     */
/* ================================================================================ */
void lu_solve_period( double L[][M+1], double U[][M], double b[], double x[] )
{
  int i;
  double sum;
  b[0]  /=  L[0][0];
  for( i = 1; i <= M-1; i++ )
    b[i]  =  (b[i]-L[1][i]*b[i-1])/L[0][i];
  sum  =  0.0;
  for( i = 0; i <= M-1; i++ )
    sum  +=  L[2][i]*b[i];
  x[M  ]  =  (b[M]-sum)/L[2][M];
  x[M-1]  =  b[M-1]-U[1][M-1]*x[M];
  for( i = M-2; i >= 0; i-- )
    x[i]  =  b[i]-U[0][i]*x[i+1]-U[1][i]*x[M];
}
/* ================================================================================ */
/*           Eӂ̐ݒ                                                             */
/* ================================================================================ */
void rightside_period( double d, double u[], double f[], double RS[] )
{
  int i;
  i  =  0;                    RS[i] = u[i] + d*re*(u[i-1+N]-2.0*u[i]+u[i+1]) + dt*f[i];
  for( i = 1; i <= M; i++ )   RS[i] = u[i] + d*re*(u[i-1  ]-2.0*u[i]+u[i+1]) + dt*f[i];
}
/* ================================================================================ */
/*           LU                                                                 */
/* ================================================================================ */
void lu_decomp_period( double a[], double b[], double c[], double L[][M+1], double U[][M] )
{
  int    i;
  double sum;
  L[0][0]  =  a[0];
  U[0][0]  =  c[0] / L[0][0];
  U[1][0]  =  b[0] / L[0][0];
  for( i = 1; i<= M-2; i++ ){
    L[1][i]  =  b[i];
    L[0][i]  =  a[i]-U[0][i-1]*L[1][i];
    U[0][i]  =  c[i]/L[0][i];
    U[1][i]  =  -L[1][i]*U[1][i-1]/L[0][i];
  }
  i        =  M-1;
  L[1][i]  =  b[i];
  L[0][i]  =  a[i]-U[0][i-1]*L[1][i];
  U[1][i]  =  (c[i]-L[1][i]*U[1][i-1])/L[0][i];
  L[2][0]  =  c[M];
  for( i = 1; i<= M-2; i++ )
    L[2][i]  =  -U[0][i-1]*L[2][i-1];
  i  =  M-1;
  L[2][i]  =  b[M]-U[0][i-1]*L[2][i-1];
  sum  =  0.0;
  for( i = 0; i<= M-1; i++ )
    sum  +=  L[2][i]*U[1][i];
  L[2][M]  =  a[M]-sum;
}
/* ================================================================================ */
/*           s̐ݒ                                                             */
/* ================================================================================ */
void matrix_period( double d, double a[], double b[], double c[] )
{
  int i;
  i  =  0;                       b[i] = -d*ri;  a[i] = 1.0+2.0*d*ri;  c[i] = -d*ri;
  for( i = 1; i <= M-1; i++ ){   b[i] = -d*ri;  a[i] = 1.0+2.0*d*ri;  c[i] = -d*ri;   }
  i  =  M;                       b[i] = -d*ri;  a[i] = 1.0+2.0*d*ri;  c[i] = -d*ri;
}


/* ================================================================================ */
/*           main֐                                                               */
/* ================================================================================ */
int main()
{
  int    step;
  FILE   *fp, *gp;
  double a[M+1], b[M+1], c[M+1]; 
  double u[N+1], f[M+1], RSu[M+1];
  double v[N+1], g[M+1], RSv[M+1];
  double Lu[3][M+1], Uu[2][M];
  double Lv[3][M+1], Uv[2][M];
  make_dir();                                                /*  f[^fBNg̍쐬               */
  step  =  0;                                                /*                                         */
  initial_condition( u, v );                                 /*  l̐ݒ                           */
  fp  =  output_gnuplot_ini();                               /*                                         */
  output( step/SOUTPUT, step*dt, u, v, fp );                 /*  f[^̏o                           */
  gp  =  gnuplot_ini();                                      /*  gnuplot̏ݒ                      */
  gnuplot( step/SOUTPUT, gp );                               /*  gnuplot̏o                          */
  usleep( SLEEP*1.0E+06 );                                   /*  SLEEPb~                  */
  matrix_period   ( du, a, b, c      );                      /*  s̐ݒ(u)                          */
  lu_decomp_period( a , b, c, Lu, Uu );                      /*  LU    (u)                          */
  matrix_period   ( dv, a, b, c      );                      /*  s̐ݒ(v)                          */
  lu_decomp_period( a , b, c, Lv, Uv );                      /*  LU    (v)                          */
  for( step = 1; step <= MAXSTEP; step++ ){                  /*                                         */
    reaction_term   ( u , v , f  , g   );                    /*  ̐ݒ                           */
    rightside_period( du, u , f  , RSu );                    /*  Eӂ̐ݒ(u)                          */
    rightside_period( dv, v , g  , RSv );                    /*  Eӂ̐ݒ(v)                          */
    lu_solve_period ( Lu, Uu, RSu, u   );                    /*  LUpĘAP莮(u)  */
    lu_solve_period ( Lv, Uv, RSv, v   );                    /*  LUpĘAP莮(v)  */
    u[N]  =  u[0];                                           /*  E(u)                        */
    v[N]  =  v[0];                                           /*  E(v)                        */
    if(  step%SOUTPUT == 0  ){                               /*                                         */
      output ( step/SOUTPUT, step*dt, u, v, fp );            /*  f[^̏o                           */
      gnuplot( step/SOUTPUT, gp                );            /*  gnuplot̏o                          */
      usleep( SLEEP*1.0E+06 );                               /*  SLEEPb~                  */
    }                                                        /*                                         */
  }                                                          /*                                         */
  printf( "please hit return key!!" );                       /*                                         */
  fprintf( fp, "!echo -n \"please hit return key!!\"\n" );   /*                                         */
  fprintf( fp, "pause -1"                               );   /*                                         */
  getchar();                                                 /*  W͂Pǂݍ           */
  fclose( fp );                                              /*  t@C                       */
  pclose( gp );                                              /*  gnuplot                        */
  return 0;                                                  /*  I                               */
}                                                            /*                                         */
