/* ================================================================================ */
/* |                                                                              | */
/* |         Mz@ŉvO - neumann -                         | */
/* |                                                                              | */
/* |         u_{t}  =  du * (u_{xx}+u_{yy})                                       | */
/* |                                                                              | */
/* |                                       This program is in the public domain.  | */
/* |                                Programmed by M.Yadome. 2008/12/18, ver. 1.0  | */
/* |                                                                              | */
/* ================================================================================ */


/* ================================================================================ */
/*           wb_t@C                                                         */
/* ================================================================================ */
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include "GL/glut.h"


/* ================================================================================ */
/*           }N`                                                             */
/* ================================================================================ */
#define  DATADIR     "data"             /*  f[^fBNg                    */ 
#define  DIRMODE     0744               /*  f[^fBNg̋[h      */
#define  MAXSTEP     3000               /*  ővZ                            */
#define  SOUTPUT     100                /*  SOUTPUTƂɃf[^o                 */
#define  SLEEP       0.0                /*  SLEEPb~                   */
#define  dt          0.0002             /*  ԍݕ                              */
#define  Nx          100                /*  x̋ԕ                       */
#define  Ny          100                /*  y̋ԕ                       */
#define  Lx          M_PI               /*  x̋                             */
#define  Ly          M_PI               /*  y̋                             */
#define  du          1.0                /*  MgU                                */
#define  BCL         0.0                /*  E()                            */
#define  BCR         0.0                /*  E(E)                            */
#define  BCB         0.0                /*  E()                            */
#define  BCT         0.0                /*  E()                            */
#define  dx          Lx/Nx
#define  dy          Ly/Ny
#define  rdx         du*dt/(dx*dx)
#define  rdy         du*dt/(dy*dy)
#define explicit( i1, i2, j3, j4 )  \
   u[i][j] + rdx*(u[i-1+i1][j     ]-2.0*u[i][j]+u[i+1+i2][j      ]) \
           + rdy*(u[i     ][j-1+j3]-2.0*u[i][j]+u[i     ][j+1+j4])
#define  WIDTH       256
#define  HEIGHT      256
#define  SCALEX      0.9
#define  SCALEY      0.9
#define  IMAGEDIR    "image"
#define  IMAGENAMEU  "u"


/* ================================================================================ */
/*           f[^fBNg̍쐬                                               */
/* ================================================================================ */
void make_dir()
{
  char str[100];
  if(  mkdir(DATADIR,DIRMODE) != 0  ){
    sprintf( str, "rm -rf %s", DATADIR );
    system( str );
    mkdir( DATADIR, DIRMODE );
  }
  if(  mkdir(IMAGEDIR,DIRMODE) != 0  ){
    sprintf( str, "rm -rf %s", IMAGEDIR );
    system( str );
    mkdir( IMAGEDIR, DIRMODE );
  }
}


/* ================================================================================ */
/*           l̐ݒ                                                           */
/* ================================================================================ */
void initial_condition( double u[][Ny+1] )
{
  int i, j;
  for( i = 0; i <= Nx; i++ )
    for( j = 0; j <= Ny; j++ )
      u[i][j]  =  ( cos(  i*M_PI/Nx)+cos(  j*M_PI/Ny) \
               +    cos(3*i*M_PI/Nx)+cos(3*j*M_PI/Ny) \
               +    cos(7*i*M_PI/Nx)+cos(7*j*M_PI/Ny) ) /4.0;
}


/* ================================================================================ */
/*                                                                              */
/* ================================================================================ */
int digit_number( int num )
{
  int dn;
  dn  =  1;
  while( (num=num/10)>0 ){   dn++;   }
  return dn;
}


/* ================================================================================ */
/*           t@C݃[hŊJ                                         */
/* ================================================================================ */
FILE *fopen_w( char filename[] )
{
  FILE *fp;
  if(  (fp=fopen(filename,"w")) == NULL  ){
    printf( "cannot open output file\n" );
    exit(1);
  }
  return fp;
}


/* ================================================================================ */
/*           f[^̏o                                                           */
/* ================================================================================ */
FILE *output_gnuplot_ini()
{
  char filename[100];
  FILE *fp;
  sprintf( filename, "%s/splot", DATADIR );
  fp  =  fopen_w( filename );
  fprintf( fp, "set xrange[0:%20.15lf]\n", Lx );
  fprintf( fp, "set yrange[0:%20.15lf]\n", Ly );
  fprintf( fp, "set zrange[-1.0:1.0]\n"       );  
  fprintf( fp, "set contour base\n"           );
  return fp;
}
void output_gnuplot( int number, FILE *fp )
{
  fprintf( fp, "!echo \"%0*d\"\n"            , digit_number(MAXSTEP/SOUTPUT), number );
  fprintf( fp, "splot \"%0*d\" u 2:3:4 w l\n", digit_number(MAXSTEP/SOUTPUT), number );
  fprintf( fp, "!sleep %lf\n"                , SLEEP                                 );
  fflush( fp );
}
void output_data( double time, double u[][Ny+1], FILE *fp )
{
  int i, j;
  for( i = 0; i <= Nx; i++ ){
    for( j = 0; j <= Ny; j++ )
      fprintf( fp, "%20.15E %20.15E %20.15E %20.15E\n", time, i*dx, j*dy, u[i][j] );
    fprintf( fp, "\n" );
  }
  fflush( fp );
}
void output( int number, double time, double u[][Ny+1], FILE *fp_gnuplot )
{
  char filename[100];
  FILE *fp_data;
  sprintf( filename, "%s/%0*d", DATADIR, digit_number(MAXSTEP/SOUTPUT), number );
  printf( "%s\n", filename );
  fp_data  =  fopen_w( filename );
  output_data( time, u, fp_data );
  fclose( fp_data );
  output_gnuplot( number, fp_gnuplot );
}


/* ================================================================================ */
/*           gnuplot̏ݒƏo                                                */
/* ================================================================================ */
FILE *gnuplot_ini()
{
  FILE *fp;
  fp  =  popen( "gnuplot -geometry 400x400+40+50", "w" );
  fprintf( fp, "set xrange[0:%20.15lf]\n", Lx );
  fprintf( fp, "set yrange[0:%20.15lf]\n", Ly );
  fprintf( fp, "set zrange[-1.0:1.0]\n"       );  
  fprintf( fp, "set hidden3d\n"               );
  fprintf( fp, "set contour base\n"           );
  return fp;
}
void gnuplot( int number, FILE *gp )
{
  fprintf( gp, "splot \"%s/%0*d\" u 2:3:4 w l\n", DATADIR, digit_number(MAXSTEP/SOUTPUT), number );
  fflush( gp );
}


/* ================================================================================ */
/*           z@                                                                 */
/* ================================================================================ */
void explicit_neumann( double u[][Ny+1] )
{
  int    i, j;
  double l, r, b, t;
  double new_u[Nx+1][Ny+1];
  l   =  - rdx*2.0*BCL*dx;
  r   =    rdx*2.0*BCR*dx;
  b   =  - rdy*2.0*BCB*dy;
  t   =    rdy*2.0*BCT*dy;
       i = 0;                          j = 0;                    new_u[i][j] = explicit(  2 ,  0 ,  2 ,  0  )+l+b;
       i = 0;                     for( j = 1; j <= Ny-1; j++ )   new_u[i][j] = explicit(  2 ,  0 ,  0 ,  0  )+l;
       i = 0;                          j = Ny;                   new_u[i][j] = explicit(  2 ,  0 ,  0 , -2  )+l+t;
  for( i = 1; i <= Nx-1; i++ ){        j = 0;                    new_u[i][j] = explicit(  0 ,  0 ,  2 ,  0  )+b;     }
  for( i = 1; i <= Nx-1; i++ )    for( j = 1; j <= Ny-1; j++ )   new_u[i][j] = explicit(  0 ,  0 ,  0 ,  0  );
  for( i = 1; i <= Nx-1; i++ ){        j = Ny;                   new_u[i][j] = explicit(  0 ,  0 ,  0 , -2  )+t;     }
       i = Nx;                         j = 0;                    new_u[i][j] = explicit(  0 , -2 ,  2 ,  0  )+r+b;
       i = Nx;                    for( j = 1; j <= Ny-1; j++ )   new_u[i][j] = explicit(  0 , -2 ,  0 ,  0  )+r;
       i = Nx;                         j = Ny;                   new_u[i][j] = explicit(  0 , -2 ,  0 , -2  )+r+t;
  for( i = 0; i <= Nx  ; i++ )    for( j = 0; j <= Ny  ; j++ )       u[i][j] = new_u[i][j];
}


/* ================================================================================ */
/*           OpenGL̏ݒƏo                                                 */
/* ================================================================================ */
void change_image_format( char *in, char *out )
{
   char string[200];
   sprintf( string, "convert %s %s", in, out );
   system( string );
}
void save_file( double width, double height, FILE *fp )
{
  int       i, x, y;
  const int w=(int)ceil(width );
  const int h=(int)ceil(height);
  char buffer[w*h*4];
  glReadBuffer( GL_FRONT );
  glReadBuffer( GL_BACK  );
  glReadPixels( 0, 0, width, height, GL_RGBA, GL_UNSIGNED_BYTE, buffer );
  fprintf( fp, "P6\n"                      );
  fprintf( fp, "# ppm-file created\n"      );
  fprintf( fp, "%i %i\n"            , w, h );
  fprintf( fp, "255\n"                     );
  for( y = height-1; y >= 0; y-- ){
    for( x = 0; x < width; x++ ){
      i  =  (y*width+x)*4;
      fputc( buffer[i  ], fp );
      fputc( buffer[i+1], fp );
      fputc( buffer[i+2], fp );
    }
  }
}
void draw( double width, double height, double R[][Ny+1], double G[][Ny+1], double B[][Ny+1] )
{
  int    i , j;
  double DX, DY;
  double sx, sy, sz;
  DX  =  width  / Nx;
  DY  =  height / Ny;
  sx  =  SCALEX;
  sy  =  SCALEY;
  sz  =  1.0;
  glClear( GL_COLOR_BUFFER_BIT );
  glTranslated( (width-sx*width)/2.0, (height-sy*height)/2.0, 0.0 );
  glScaled( sx, sy, sz );
  for( i = 0; i < Nx; i++ ){
    for( j = 0; j < Ny; j++ ){
      glBegin( GL_POLYGON );
        glColor3f( R[i  ][j  ], G[i  ][j  ], B[i  ][j  ] );   glVertex2f(  i   *DX,  j   *DY );
        glColor3f( R[i+1][j  ], G[i+1][j  ], B[i+1][j  ] );   glVertex2f( (i+1)*DX,  j   *DY );
        glColor3f( R[i+1][j+1], G[i+1][j+1], B[i+1][j+1] );   glVertex2f( (i+1)*DX, (j+1)*DY );
        glColor3f( R[i  ][j+1], G[i  ][j+1], B[i  ][j+1] );   glVertex2f(  i   *DX, (j+1)*DY );
      glEnd();
    }
  }
}
void display( double width, double height, double R[][Ny+1], double G[][Ny+1], double B[][Ny+1] )
{
  glViewport( 0, 0, width, height );
    glMatrixMode( GL_PROJECTION );
    glLoadIdentity();
    glOrtho( 0, width, 0, height, -1.0, 1.0 );
      glMatrixMode( GL_MODELVIEW );
      glLoadIdentity();
      draw( width, height, R, G, B );
      glPopMatrix();
    glPopMatrix();
  glutSwapBuffers();
}
void  max_min( double u[][Ny+1], double *max, double *min )
{
  int i, j;
  *max  =  u[0][0];
  *min  =  u[0][0];
  for( i = 0; i <= Nx; i++ ){
    for( j = 0; j <= Ny; j++ ){
      if(  *max < u[i][j]  )   *max = u[i][j];
      if(  *min > u[i][j]  )   *min = u[i][j];
    }
  }
}
void color( double max_x, double min_x, double u[][Ny+1], double R[][Ny+1], double G[][Ny+1], double B[][Ny+1] )
{
  int    i, j;
  double color;
  if(  max_x == min_x  ){
    for( i = 0; i <= Nx; i++ ){
      for( j = 0; j <= Ny; j++ ){
        R[i][j]  =  0.0;
        G[i][j]  =  0.0;
        B[i][j]  =  0.0;
      }
    }
  }else{
    for( i = 0; i <= Nx; i++ ){
      for( j = 0; j <= Ny; j++ ){
        color  =  (u[i][j]-min_x)/(max_x-min_x);
        if     (   color <  0.0                      ){   R[i][j] = 0.0;             G[i][j] = 0.0;             B[i][j] = 1.0;             }
        else if(   color >= 1.0                      ){   R[i][j] = 1.0;             G[i][j] = 0.0;             B[i][j] = 0.0;             }
        else if(  (color >= 0.00) && (color < 0.25)  ){   R[i][j] = 0.0;             G[i][j] = 4.0*color;       B[i][j] = 1.0;             }
        else if(  (color >= 0.25) && (color < 0.50)  ){   R[i][j] = 0.0;             G[i][j] = 1.0;             B[i][j] = 2.0-4.0*color;   }
        else if(  (color >= 0.50) && (color < 0.75)  ){   R[i][j] = 4.0*color-2.0;   G[i][j] = 1.0;             B[i][j] = 0.0;             }
        else if(  (color >= 0.75) && (color < 1.00)  ){   R[i][j] = 1.0;             G[i][j] = 4.0-4.0*color;   B[i][j] = 0.0;             }
      }
    }
  }
}
void image( int number, int handle, char *name, double max_x, double min_x, double u[][Ny+1] )
{
  char   ppmfile[100], jpegfile[100];
  FILE   *fp;
  double width, height;
  double R[Nx+1][Ny+1], G[Nx+1][Ny+1], B[Nx+1][Ny+1];
  glutSetWindow( handle );
  width   =  (double)glutGet( GLUT_WINDOW_WIDTH  );
  height  =  (double)glutGet( GLUT_WINDOW_HEIGHT );
  color( max_x, min_x, u, R, G, B );
  display( width, height, R, G, B );
  sprintf( ppmfile , "%s/%s%0*d.ppm", IMAGEDIR, name, digit_number(MAXSTEP/SOUTPUT), number );
  sprintf( jpegfile, "%s/%s%0*d.jpg", IMAGEDIR, name, digit_number(MAXSTEP/SOUTPUT), number );
  fp  =  fopen_w( ppmfile );
  save_file( width, height, fp );
  fclose( fp );
  change_image_format( ppmfile, jpegfile );
  unlink( ppmfile );
}
void image_ini( int px, int py, char *handle_name, int *handle )
{
  glutInitDisplayMode( GLUT_RGBA | GLUT_DOUBLE );
  glutInitWindowSize( WIDTH, HEIGHT );
  glutInitWindowPosition( px, py );
  *handle  =  glutCreateWindow( handle_name );
  glClearColor( 1.0, 1.0, 1.0, 1.0 );
}


/* ================================================================================ */
/*           main֐                                                               */
/* ================================================================================ */
int main( int argc, char *argv[] )
{
  int    step;
  int    handle_u;
  FILE   *fp, *gp;
  double max_u, min_u;
  double u[Nx+1][Ny+1];
  make_dir();                                                             /*  f[^fBNg̍쐬      */
  step  =  0;                                                             /*                                */
  initial_condition( u );                                                 /*  l̐ݒ                  */
  fp  =  output_gnuplot_ini();                                            /*                                */
  output( step/SOUTPUT, step*dt, u, fp );                                 /*  f[^̏o                  */
  gp  =  gnuplot_ini();                                                   /*  gnuplot̏ݒ             */ 
  gnuplot  ( step/SOUTPUT , gp                                     );     /*  gnuplot̏o                 */
  glutInit ( &argc, argv                                           );     /*  OpenGL̏ݒ              */
  image_ini( 450  , 50    ,"Solution", &handle_u                   );     /*  OpenGL̏ݒ              */
  max_min  ( u    , &max_u, &min_u                                 );     /*  őlƍŏľvZ          */
  image    ( step/SOUTPUT , handle_u , IMAGENAMEU, max_u, min_u, u );     /*  OpenGL̏o                  */
  usleep( SLEEP*1.0E+06 );                                                /*  SLEEPb~         */
  for( step = 1; step <= MAXSTEP; step++ ){                               /*                                */
    explicit_neumann( u );                                                /*  z@                        */
    if(  step%SOUTPUT == 0  ){                                            /*                                */
      output ( step/SOUTPUT, step*dt , u         , fp              );     /*  f[^̏o                  */
      gnuplot( step/SOUTPUT, gp                                    );     /*  gnuplot̏o                 */
//      max_min( u           , &max_u  , &min_u                      );     /*  őlƍŏľvZ          */
      image  ( step/SOUTPUT, handle_u, IMAGENAMEU, max_u, min_u, u );     /*  OpenGL̏o                  */
      usleep( SLEEP*1.0E+06 );                                            /*  SLEEPb~         */
    }                                                                     /*                                */
  }                                                                       /*                                */
  printf (     "please hit return key!!"                );                /*                                */
  fprintf( fp, "!echo -n \"please hit return key!!\"\n" );                /*                                */
  fprintf( fp, "pause -1"                               );                /*                                */
  getchar();                                                              /*  W͂Pǂݍ  */
  fclose( fp );                                                           /*  t@C              */
  pclose( gp );                                                           /*  gnuplot               */
  return 0;                                                               /*  I                      */
}
