/* ================================================================================ */
/* |                                                                              | */
/* |         gFitzHugh-NagumoADI@ŉvO - dirichlet -   | */
/* |                                                                              | */
/* |         u_{t}  =  du * (u_{xx}+u_{yy}) + (u*(1-u)*(u-a)-v) /eps              | */
/* |         v_{t}  =  dv * (v_{xx}+v_{yy}) +  x-gam*y                            | */
/* |                                                                              | */
/* |                                       This program is in the public domain.  | */
/* |                                Programmed by M.Yadome. 2008/12/18, ver. 1.0  | */
/* |                                                                              | */
/* ================================================================================ */


/* ================================================================================ */
/*           wb_t@C                                                         */
/* ================================================================================ */
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include "GL/glut.h"


/* ================================================================================ */
/*           }N`                                                             */
/* ================================================================================ */
#define  DATADIR     "data"             /*  f[^fBNg                    */
#define  DIRMODE     0744               /*  f[^fBNg̋[h      */
#define  MAXSTEP     3000               /*  ővZ                            */
#define  SOUTPUT     100                /*  SOUTPUTƂɃf[^o                 */
#define  SLEEP       0.0                /*  SLEEPb~                   */
#define  dt          0.001              /*  ԍݕ                              */
#define  Nx          100                /*  x̋ԕ                       */
#define  Ny          100                /*  y̋ԕ                       */
#define  Lx          20.0               /*  x̋                             */
#define  Ly          20.0               /*  y̋                             */
#define  du          1.0                /*  gUW                                */
#define  dv          0.1                /*  gUW                                */
#define  eps         0.001              /*  ̃p[^                      */
#define  aa          0.125              /*  ̃p[^                      */
#define  gam         1.0                /*  ̃p[^                      */
#define  BCL         0.0                /*  E()                            */
#define  BCR         0.0                /*  E(E)                            */
#define  BCB         0.0                /*  E()                            */
#define  BCT         0.0                /*  E()                            */
#define  dx          Lx/Nx
#define  dy          Ly/Ny
#define  ri1         0.5*dt/(dx*dx)
#define  re1         0.5*dt/(dy*dy)
#define  ri2         re1
#define  re2         ri1
#define  Mx          Nx-1
#define  My          Ny-1
#define  bcil        -ri1*BCL
#define  bcir        ri1*BCR
#define  bcib        -ri2*BCB
#define  bcit        ri2*BCT
#define  adi_explicit_x( rd, i1, i2 )  u[i][j] + rd*(u[i-1+i1][j     ]-2.0*u[i][j]+u[i+1+i2][j     ]) + 0.5*dt*f[i][j]
#define  adi_explicit_y( rd, j1, j2 )  u[i][j] + rd*(u[i     ][j-1+j1]-2.0*u[i][j]+u[i     ][j+1+j2]) + 0.5*dt*f[i][j]
#define  WIDTH       256
#define  HEIGHT      256
#define  SCALEX      0.9
#define  SCALEY      0.9
#define  IMAGEDIR    "image"
#define  IMAGENAMEU  "u"
#define  IMAGENAMEV  "v"


/* ================================================================================ */
/*           f[^fBNg̍쐬                                               */
/* ================================================================================ */
void make_dir()
{
  char str[100];
  if(  mkdir(DATADIR,DIRMODE) != 0  ){
    sprintf( str, "rm -rf %s", DATADIR );
    system( str );
    mkdir( DATADIR, DIRMODE );
  }
  if(  mkdir(IMAGEDIR,DIRMODE) != 0  ){
    sprintf( str, "rm -rf %s", IMAGEDIR );
    system( str );
    mkdir( IMAGEDIR, DIRMODE );
  }
}


/* ================================================================================ */
/*           t@Cǂݍ݃[hŊJ                                         */
/* ================================================================================ */
FILE *fopen_r( char filename[] )
{
  FILE *fp;
  if(  (fp=fopen(filename,"r")) == NULL  ){
    printf( "cannot open output file\n" );
    exit(1);
  }
  return fp;
}


/* ================================================================================ */
/*           l̐ݒ                                                           */
/* ================================================================================ */
void initial_condition( double u[][Ny+1], double v[][Ny+1] )
{
  int i, j;
  FILE   *fp;
  double time, idx;
  fp  =  fopen_r( "tp.dat" );
  for( i = 0; i <= Nx; i++ )
    for( j = 0; j <= Ny; j++ )
      u[i][j]  =  0.0;
  for( i = 0; i <= Nx; i++ )
    fscanf( fp, "%lf %lf %lf %lf\n", &time, &idx, &u[i][0], &v[i][0] );
  fclose( fp );
  for( j = 1; j <= Ny/5; j++ )
    for( i = 0; i <= Nx; i++ ){
      u[i][j]  =  u[i][0];
      v[i][j]  =  v[i][0];
    }
}


/* ================================================================================ */
/*           FitzHugh-Nagumo equation                                               */
/* ================================================================================ */
void reaction_term( double u[][Ny+1], double v[][Ny+1], double f[][My+1], double g[][My+1] )
{
  int i, j;
  for( i = 0; i <= Mx; i++ ){
    for( j = 0; j <= My; j++ ){
      f[i][j]  =  (u[i][j]*(1.0-u[i][j])*(u[i][j]-aa)-v[i][j])/eps;
      g[i][j]  =  u[i][j]-gam*v[i][j];
    }
  }
}


/* ================================================================================ */
/*                                                                              */
/* ================================================================================ */
int digit_number( int num )
{
  int dn;
  dn  =  1;
  while( (num=num/10)>0 ){   dn++;   }
  return dn;
}


/* ================================================================================ */
/*           t@C݃[hŊJ                                         */
/* ================================================================================ */
FILE *fopen_w( char filename[] )
{
  FILE *fp;
  if(  (fp=fopen(filename,"w")) == NULL  ){
    printf( "cannot open output file\n" );
    exit(1);
  }
  return fp;
}


/* ================================================================================ */
/*           f[^̏o                                                           */
/* ================================================================================ */
FILE *output_gnuplot_ini()
{
  char filename[100];
  FILE *fp;
  sprintf( filename, "%s/splot", DATADIR );
  fp  =  fopen_w( filename );
  fprintf( fp, "set xrange[0:%20.15lf]\n", Lx );
  fprintf( fp, "set yrange[0:%20.15lf]\n", Ly );
  fprintf( fp, "set zrange[-1.0:1.0]\n"       );  
  fprintf( fp, "set contour base\n"           );
  return fp;
}
void output_gnuplot( int number, FILE *fp )
{
  fprintf( fp, "!echo \"%0*d\"\n"            , digit_number(MAXSTEP/SOUTPUT), number );
  fprintf( fp, "splot \"%0*d\" u 2:3:4 w l\n", digit_number(MAXSTEP/SOUTPUT), number );
  fprintf( fp, "!sleep %lf\n"                , SLEEP                                 );
  fflush( fp );
}
void output_data( double time, double u[][Ny+1], double v[][Ny+1], FILE *fp )
{
  int i, j;
  for( i = 0; i <= Nx; i++ ){
    for( j = 0; j <= Ny; j++ )
      fprintf( fp, "%20.15E %20.15E %20.15E %20.15E %20.15E\n", time, i*dx, j*dy, u[i][j], v[i][j] );
    fprintf( fp, "\n" );
  }
  fflush( fp );
}
void output( int number, double time, double u[][Ny+1], double v[][Ny+1], FILE *fp_gnuplot )
{
  char filename[100];
  FILE *fp_data;
  sprintf( filename, "%s/%0*d", DATADIR, digit_number(MAXSTEP/SOUTPUT), number );
  printf( "%s\n", filename );
  fp_data  =  fopen_w( filename    );
  output_data( time, u, v, fp_data );
  fclose( fp_data );
  output_gnuplot( number, fp_gnuplot );
}


/* ================================================================================ */
/*           gnuplot̏ݒƏo                                                */
/* ================================================================================ */
FILE *gnuplot_ini()
{
  FILE *fp;
  fp  =  popen( "gnuplot -geometry 400x400+40+50", "w" );
  fprintf( fp, "set xrange[0:%20.15lf]\n", Lx );
  fprintf( fp, "set yrange[0:%20.15lf]\n", Ly );
  fprintf( fp, "set zrange[-1.0:1.0]\n"       );  
  fprintf( fp, "set hidden3d\n"               );
  fprintf( fp, "set contour base\n"           );
  return fp;
}
void gnuplot( int number, FILE *gp )
{
  fprintf( gp, "splot \"%s/%0*d\" u 2:3:4 w l\n", DATADIR, digit_number(MAXSTEP/SOUTPUT), number );
  fflush( gp );
}


/* ================================================================================ */
/*           ̍XV                                                               */
/* ================================================================================ */
void update_sol_dirichlet( double u[][Ny+1], double new_u[][My+1] )
{
  int i, j;
       i = 0;                  for( j = 0; j <= Ny; j++ )   u[i][j] = BCL;
       i = Nx;                 for( j = 0; j <= Ny; j++ )   u[i][j] = BCR;
       j = 0;                  for( i = 1; i <= Mx; i++ )   u[i][j] = BCB;
       j = Ny;                 for( i = 1; i <= Mx; i++ )   u[i][j] = BCT;
  for( i = 1; i <= Mx; i++ )   for( j = 1; j <= My; j++ )   u[i][j] = new_u[i][j];
}
/* ================================================================================ */
/*           LUpĘAP莮                                     */
/* ================================================================================ */
void lu_solve_dirichlet( const int M, double L[][M+1], double U[], double b[], double x[] )
{
  int i;
  b[1]  /=  L[0][1];
  for( i = 2; i <= M; i++ )
    b[i]  =  (b[i]-L[1][i]*b[i-1])/L[0][i];
  x[M]  =  b[M];
  for( i = M-1; i >= 1; i-- )
    x[i]  =  b[i]-U[i]*x[i+1];
}
/* ================================================================================ */
/*           Eӂ̐ݒ                                                             */
/* ================================================================================ */
void rightside_xM_dirichlet( int i, int M, double d, double r, double u[][Ny+1], double f[][My+1], double RS[] )
{    int j;  for( j = 1; j <= M; j++ ) RS[j] = adi_explicit_x( d*r, 0, 0 );   RS[1]+=d*bcib;   RS[M]+=d*bcit;    }
void rightside_yM_dirichlet( int j, int M, double d, double r, double u[][Ny+1], double f[][My+1], double RS[] )
{    int i;  for( i = 1; i <= M; i++ ) RS[i] = adi_explicit_y( d*r, 0, 0 );   RS[1]+=d*bcil;   RS[M]+=d*bcir;    }
/* ================================================================================ */
/*           ADI 2i                                                              */
/* ================================================================================ */
void adi2_dirichlet( double d, double r, double u[][Ny+1], double f[][My+1], double L[][My+1], double U[], double new_u[][My+1] )
{
  int    i, j;
  double RS[My+1], s[My+1];
  for( i = 1; i <= Mx-1; i++ ){
    rightside_xM_dirichlet( i , My, d, r , u, f, RS );   for( j = 0; j <= My; j++ ) s       [j] = u [i][j];
    lu_solve_dirichlet    ( My, L , U, RS, s        );   for( j = 0; j <= My; j++ ) new_u[i][j] = s    [j];
  }
}
/* ================================================================================ */
/*           ADI 1i                                                              */
/* ================================================================================ */
void adi1_dirichlet( double d, double r, double u[][Ny+1], double f[][My+1], double L[][Mx+1], double U[], double new_u[][My+1] )
{
  int    i, j;
  double RS[Mx+1], s[Mx+1];
  for( j = 1; j <= My-1; j++ ){
    rightside_yM_dirichlet( j , Mx, d, r , u, f, RS );   for( i = 0; i <= Mx; i++ ) s    [i]    = u[i][j];
    lu_solve_dirichlet    ( Mx, L , U, RS, s        );   for( i = 0; i <= Mx; i++ ) new_u[i][j] = s[i];
  }
}
/* ================================================================================ */
/*           LU                                                                 */
/* ================================================================================ */
void lu_decomp_dirichlet( const int M, double a[], double b[], double c[], double L[][M+1], double U[] )
{
  int i;
  L[0][1]  =  a[1];
  for( i = 2; i <= M; i++ )
    L[1][i]  =  b[i];
  for( i = 1; i <= M-1; i++ ){
    U[i]       =  c[i] / L[0][i];
    L[0][i+1]  =  a[i+1]-U[i]*L[1][i+1];
  }
}
/* ================================================================================ */
/*           s̐ݒ                                                             */
/* ================================================================================ */
void matrix_dirichlet( int M, double rd, double a[], double b[], double c[] )
{
  int i;
  i  =  1;                                    a[i] = 1.0+2.0*rd;  c[i] = -rd;
  for( i = 2; i <= M-1; i++ ){   b[i] = -rd;  a[i] = 1.0+2.0*rd;  c[i] = -rd;   }
  i  =  M;                       b[i] = -rd;  a[i] = 1.0+2.0*rd;
}


/* ================================================================================ */
/*           OpenGL̏ݒƏo                                                 */
/* ================================================================================ */
void change_image_format( char *in, char *out )
{
   char string[200];
   sprintf( string, "convert %s %s", in, out );
   system( string );
}
void save_file( double width, double height, FILE *fp )
{
  int       i, x, y;
  const int w=(int)ceil(width );
  const int h=(int)ceil(height);
  char buffer[w*h*4];
  glReadBuffer( GL_FRONT );
  glReadBuffer( GL_BACK  );
  glReadPixels( 0, 0, width, height, GL_RGBA, GL_UNSIGNED_BYTE, buffer );
  fprintf( fp, "P6\n"                      );
  fprintf( fp, "# ppm-file created\n"      );
  fprintf( fp, "%i %i\n"            , w, h );
  fprintf( fp, "255\n"                     );
  for( y = height-1; y >= 0; y-- ){
    for( x = 0; x < width; x++ ){
      i  =  (y*width+x)*4;
      fputc( buffer[i  ], fp );
      fputc( buffer[i+1], fp );
      fputc( buffer[i+2], fp );
    }
  }
}
void draw( double width, double height, double R[][Ny+1], double G[][Ny+1], double B[][Ny+1] )
{
  int    i , j;
  double DX, DY;
  double sx, sy, sz;
  DX  =  width  / Nx;
  DY  =  height / Ny;
  sx  =  SCALEX;
  sy  =  SCALEY;
  sz  =  1.0;
  glClear( GL_COLOR_BUFFER_BIT );
  glTranslated( (width-sx*width)/2.0, (height-sy*height)/2.0, 0.0 );
  glScaled( sx, sy, sz );
  for( i = 0; i < Nx; i++ ){
    for( j = 0; j < Ny; j++ ){
      glBegin( GL_POLYGON );
        glColor3f( R[i  ][j  ], G[i  ][j  ], B[i  ][j  ] );   glVertex2f(  i   *DX,  j   *DY );
        glColor3f( R[i+1][j  ], G[i+1][j  ], B[i+1][j  ] );   glVertex2f( (i+1)*DX,  j   *DY );
        glColor3f( R[i+1][j+1], G[i+1][j+1], B[i+1][j+1] );   glVertex2f( (i+1)*DX, (j+1)*DY );
        glColor3f( R[i  ][j+1], G[i  ][j+1], B[i  ][j+1] );   glVertex2f(  i   *DX, (j+1)*DY );
      glEnd();
    }
  }
}
void display( double width, double height, double R[][Ny+1], double G[][Ny+1], double B[][Ny+1] )
{
  glViewport( 0, 0, width, height );
    glMatrixMode( GL_PROJECTION );
    glLoadIdentity();
    glOrtho( 0, width, 0, height, -1.0, 1.0 );
      glMatrixMode( GL_MODELVIEW );
      glLoadIdentity();
      draw( width, height, R, G, B );
      glPopMatrix();
    glPopMatrix();
  glutSwapBuffers();
}
void  max_min( double u[][Ny+1], double *max, double *min )
{
  int i, j;
  *max  =  u[0][0];
  *min  =  u[0][0];
  for( i = 0; i <= Nx; i++ ){
    for( j = 0; j <= Ny; j++ ){
      if(  *max < u[i][j]  )   *max = u[i][j];
      if(  *min > u[i][j]  )   *min = u[i][j];
    }
  }
}
void color( double max_x, double min_x, double u[][Ny+1], double R[][Ny+1], double G[][Ny+1], double B[][Ny+1] )
{
  int    i, j;
  double color;
  if(  max_x == min_x  ){
    for( i = 0; i <= Nx; i++ ){
      for( j = 0; j <= Ny; j++ ){
        R[i][j]  =  0.0;
        G[i][j]  =  0.0;
        B[i][j]  =  0.0;
      }
    }
  }else{
    for( i = 0; i <= Nx; i++ ){
      for( j = 0; j <= Ny; j++ ){
        color  =  (u[i][j]-min_x)/(max_x-min_x);
        if     (   color <  0.0                      ){   R[i][j] = 0.0;             G[i][j] = 0.0;             B[i][j] = 1.0;             }
        else if(   color >= 1.0                      ){   R[i][j] = 1.0;             G[i][j] = 0.0;             B[i][j] = 0.0;             }
        else if(  (color >= 0.00) && (color < 0.25)  ){   R[i][j] = 0.0;             G[i][j] = 4.0*color;       B[i][j] = 1.0;             }
        else if(  (color >= 0.25) && (color < 0.50)  ){   R[i][j] = 0.0;             G[i][j] = 1.0;             B[i][j] = 2.0-4.0*color;   }
        else if(  (color >= 0.50) && (color < 0.75)  ){   R[i][j] = 4.0*color-2.0;   G[i][j] = 1.0;             B[i][j] = 0.0;             }
        else if(  (color >= 0.75) && (color < 1.00)  ){   R[i][j] = 1.0;             G[i][j] = 4.0-4.0*color;   B[i][j] = 0.0;             }
      }
    }
  }
}
void image( int number, int handle, char *name, double max_x, double min_x, double u[][Ny+1] )
{
  char   ppmfile[100], jpegfile[100];
  FILE   *fp;
  double width, height;
  double R[Nx+1][Ny+1], G[Nx+1][Ny+1], B[Nx+1][Ny+1];
  glutSetWindow( handle );
  width   =  (double)glutGet( GLUT_WINDOW_WIDTH  );
  height  =  (double)glutGet( GLUT_WINDOW_HEIGHT );
  color( max_x, min_x, u, R, G, B );
  display( width, height, R, G, B );
  sprintf( ppmfile , "%s/%s%0*d.ppm", IMAGEDIR, name, digit_number(MAXSTEP/SOUTPUT), number );
  sprintf( jpegfile, "%s/%s%0*d.jpg", IMAGEDIR, name, digit_number(MAXSTEP/SOUTPUT), number );
  fp  =  fopen_w( ppmfile );
  save_file( width, height, fp );
  fclose( fp );
  change_image_format( ppmfile, jpegfile );
  unlink( ppmfile );
}
void image_ini( int px, int py, char *handle_name, int *handle )
{
  glutInitDisplayMode( GLUT_RGBA | GLUT_DOUBLE );
  glutInitWindowSize( WIDTH, HEIGHT );
  glutInitWindowPosition( px, py );
  *handle  =  glutCreateWindow( handle_name );
  glClearColor( 1.0, 1.0, 1.0, 1.0 );
}


/* ================================================================================ */
/*           main֐                                                               */
/* ================================================================================ */
int main( int argc, char *argv[] )
{
  int    step;
  int    handle_u, handle_v;
  FILE   *fp, *gp;
  double max_u, min_u;
  double max_v, min_v;
  double ax[Mx+1], bx[Mx+1], cx[Mx+1];
  double ay[Mx+1], by[Mx+1], cy[Mx+1];
  double u[Nx+1][Ny+1], f[Mx+1][My+1], new_u[Mx+1][My+1];
  double v[Nx+1][Ny+1], g[Mx+1][My+1], new_v[Mx+1][My+1];
  double Lu1[2][Mx+1], Uu1[Mx], Lu2[2][My+1], Uu2[My];
  double Lv1[2][Mx+1], Uv1[Mx], Lv2[2][My+1], Uv2[My];
  make_dir();                                                                  /*  f[^fBNg̍쐬      */
  step  =  0;                                                                  /*                                */
  initial_condition( u, v );                                                   /*  l̐ݒ                  */
  fp  =  output_gnuplot_ini();                                                 /*                                */
  output( step/SOUTPUT, step*dt, u, v, fp );                                   /*  f[^̏o                  */
  gp  =  gnuplot_ini();                                                        /*  gnuplot̏ݒ             */
  gnuplot( step/SOUTPUT, gp );                                                 /*  gnuplot̏o                 */
  glutInit( &argc, argv                                                 );     /*  OpenGL̏ݒ              */
  image_ini( 450 , 50       , "Solution_U", &handle_u                   );     /*  OpenGL̏ݒ(u)           */
  image_ini( 450 , 80+HEIGHT, "Solution_V", &handle_v                   );     /*  OpenGL̏ݒ(v)           */
  max_min  ( u   , &max_u   , &min_u                                    );     /*  őlƍŏľvZ(u)       */
  max_min  ( v   , &max_v   , &min_v                                    );     /*  őlƍŏľvZ(v)       */
  image    ( step/SOUTPUT   , handle_u    , IMAGENAMEU, max_u, min_u, u );     /*  OpenGL̏o(u)               */
  image    ( step/SOUTPUT   , handle_v    , IMAGENAMEV, max_v, min_v, v );     /*  OpenGL̏o(v)               */
  usleep( SLEEP*1.0E+06 );                                                     /*  SLEEPb~         */
  matrix_dirichlet   ( Mx, du*ri1, ax, bx , cx  );                             /*  s̐ݒ(1i:u)           */
  lu_decomp_dirichlet( Mx, ax, bx, cx, Lu1, Uu1 );                             /*  LU    (1i:u)           */
  matrix_dirichlet   ( My, du*ri2, ay, by , cy  );                             /*  s̐ݒ(2i:u)           */
  lu_decomp_dirichlet( My, ay, by, cy, Lu2, Uu2 );                             /*  LU    (2i:u)           */
  matrix_dirichlet   ( Mx, dv*ri1, ax, bx , cx  );                             /*  s̐ݒ(1i:v)           */
  lu_decomp_dirichlet( Mx, ax, bx, cx, Lv1, Uv1 );                             /*  LU    (1i:v)           */
  matrix_dirichlet   ( My, dv*ri2, ay, by , cy  );                             /*  s̐ݒ(2i:v)           */
  lu_decomp_dirichlet( My, ay, by, cy, Lv2, Uv2 );                             /*  LU    (2i:v)           */
  for( step = 1; step <= MAXSTEP; step++ ){                                    /*                                */
    reaction_term( u, v, f, g );                                               /*  ̐ݒ                  */
    adi1_dirichlet( du, re1, u, f, Lu1, Uu1, new_u );                          /*  ADI 1i(u)                  */
    adi1_dirichlet( dv, re1, v, g, Lv1, Uv1, new_v );                          /*  ADI 1i(v)                  */
    update_sol_dirichlet( u, new_u );                                          /*  ̍XV (u)                  */
    update_sol_dirichlet( v, new_v );                                          /*  ̍XV (v)                  */
    reaction_term( u, v, f, g );                                               /*  ̐ݒ                  */
    adi2_dirichlet( du, re2, u, f, Lu2, Uu2, new_u );                          /*  ADI 2i(u)                  */
    adi2_dirichlet( dv, re2, v, g, Lv2, Uv2, new_v );                          /*  ADI 2i(v)                  */
    update_sol_dirichlet( u, new_u );                                          /*  ̍XV (u)                  */
    update_sol_dirichlet( v, new_v );                                          /*  ̍XV (v)                  */
    if(  step%SOUTPUT == 0  ){                                                 /*                                */
      output ( step/SOUTPUT, step*dt , u, v      , fp              );          /*  f[^̏o                  */
      gnuplot( step/SOUTPUT, gp                                    );          /*  gnuplot̏o                 */
//      max_min( u           , &max_u  , &min_u                      );          /*  őlƍŏľvZ(u)       */
//      max_min( v           , &max_v  , &min_v                      );          /*  őlƍŏľvZ(v)       */
      image  ( step/SOUTPUT, handle_u, IMAGENAMEU, max_u, min_u, u );          /*  OpenGL̏o(u)               */
      image  ( step/SOUTPUT, handle_v, IMAGENAMEV, max_v, min_v, v );          /*  OpenGL̏o(v)               */
      usleep( SLEEP*1.0E+06 );                                                 /*  SLEEPb~         */
    }                                                                          /*                                */
  }                                                                            /*                                */
  printf (     "please hit return key!!"                );                     /*                                */
  fprintf( fp, "!echo -n \"please hit return key!!\"\n" );                     /*                                */
  fprintf( fp, "pause -1"                               );                     /*                                */
  getchar();                                                                   /*  W͂Pǂݍ  */
  fclose( fp );                                                                /*  t@C              */
  pclose( gp );                                                                /*  gnuplot               */
  return 0;                                                                    /*  I                      */
}                                                                              /*                                */
